from dataclasses import asdict
from typing import TypeVar
from marshmallow_objects import models
from typing import Any, Dict, List, Optional, Union

from zpy.logger import zL
from zpy.utils.funcs import safely_exec, safely_exec_with
from zdb.commons import process_exception
from enum import Enum
import cx_Oracle
import logging

__author__ = "Noé Cruz | contactozurckz@gmail.com"
__copyright__ = "Copyright 2021, Small APi Project"
__credits__ = ["Noé Cruz", "Zurck'z"]
__license__ = "MIT"
__version__ = "0.0.1"
__maintainer__ = "Noé Cruz"
__email__ = "contactozurckz@gmail.com"
__status__ = "Dev"

from zdb.commons import ZDatabase, ZDBConfig, ZDBPool, show_info, parser_types
from zdb.utils import get_current_schema

T = TypeVar("T")


class OracleType(Enum):
    cursor = cx_Oracle.CURSOR
    number = cx_Oracle.NUMBER
    string = cx_Oracle.STRING
    integer = cx_Oracle.NUMBER
    decimal = cx_Oracle.NUMBER
    bool = cx_Oracle.BOOLEAN


class OracleParam(Enum):
    LIST_INTEGER = "LIST_INTEGER"
    LIST_STR = "LIST_VARCHAR"
    LIST_CLOB = "LIST_CLOB"


class ZParam:
    def __init__(
            self,
            value: Union[List[int], List[float], List[str], List[Any]],
            param_type: OracleParam,
            key: str,
            origin: str = None,
    ) -> None:
        self.value = value
        self.param_type = param_type
        self.key = key
        self.origin = origin


class IntList(ZParam):
    def __init__(self, value: List[int], key: str, origin: str = None) -> None:
        super().__init__(value, OracleParam.LIST_INTEGER, key, origin)


class StrList(ZParam):
    def __init__(self, value: List[str], key: str, origin: str = None) -> None:
        super().__init__(value, OracleParam.LIST_STR, key, origin)


class ClobList(ZParam):
    def __init__(self, value: List[Any], key: str, origin: str) -> None:
        super().__init__(value, OracleParam.LIST_CLOB, key, origin)


def get_str_connection(config: dict = None, mode="TSN") -> Union[dict, str]:
    if config is None:
        raise Exception("The data for the connection was not provided")
    server = config["host"]
    port = config["port"]
    service = config["service"]
    user = config["user"]
    password = config["password"]
    if mode == "DSN":
        return "{0}:{1}/{2}".format(server, port, service)
    return "{0}/{1}@{2}:{3}/{4}".format(user, password, server, port, service)


def extract_of_result_data(
        result_set, ret_type: OracleType, model=None, jsonfy=False
):
    """
    New version of result set processor
    """
    if ret_type == OracleType.cursor:
        columns = [field[0] for field in result_set.description]
        if model is not None:
            return [model(**dict(zip(columns, r))) for r in result_set]
        if jsonfy is True:
            return [dict(zip(columns, row)) for row in result_set]
        return result_set.fetchall()
    elif OracleType.number:
        return safely_exec_with(lambda x: float(result_set), result_set, args=[result_set])
    elif OracleType.integer:
        return safely_exec_with(lambda x: int(result_set), result_set, args=[result_set])
    elif OracleType.decimal:
        return safely_exec_with(lambda x: parser_types[4](x), result_set, args=[result_set])
    else:
        return parser_types[2](result_set)


def make_custom_param(
        connection: Any,
        param_type: OracleParam,
        value: Union[List[int], List[float], List[str], List[Any]],
        schema: str = None,
):
    """
    Make custom param
    """
    db_schema = (
        "" if (schema is None or schema.replace(" ", "") == "") else f"{schema}."
    )
    list_type = connection.gettype(f"{db_schema}{param_type.value}")
    return list_type.newobject(value)


class ZOracle(ZDatabase):
    __local_client_initialized: bool = False
    __local_client_path: str = None
    __config_connection: dict = None
    __connection = None
    __is_connected: bool = False

    __schemas: List[Dict] = None
    __env: str = None
    __verbose: bool = False

    def __init__(
            self,
            config: dict = None,
            local_client_path: str = None,
            schemas: List[Dict] = None,
            env: str = None,
            verbose: bool = False,
    ) -> None:
        self.__local_client_path = local_client_path
        self.__config_connection = config
        self.__schemas = schemas
        self.__env = env
        self.__verbose = verbose

    @classmethod
    def setup(cls, config: dict,
              local_client_path: str = None,
              schemas: List[Dict] = None,
              env: str = None,
              verbose: bool = False):
        return cls(config, local_client_path, schemas, env, verbose)

    @classmethod
    def setup_of(cls, config: ZDBConfig, local_client_path: str = None,
                 schemas: List[Dict] = None,
                 env: str = None,
                 verbose: bool = False, ):
        return cls(asdict(config), local_client_path, schemas, env, verbose)

    @classmethod
    def from_of(cls, user: str, password: str, host: str, db_name: str, verbose: bool = False):
        raise NotImplementedError("Not implemented for Oracle!")

    def new_connect(self) -> Any:
        str_connection = self.__validate_config(self.__config_connection)
        return cx_Oracle.connect(str_connection)

    def init_local_client(self, path: str = None):
        if self.__local_client_initialized:
            return
        value = path if self.__local_client_path is None else self.__local_client_path
        try:
            if value is None:
                raise Exception("Local client path not provided.")
            cx_Oracle.init_oracle_client(lib_dir=value)
            self.__local_client_initialized = True
        except Exception as e:
            self.__local_client_initialized = False
            logging.exception(e)

    def __validate_config(self, config: dict = None, mode="TSN") -> Union[dict, str]:
        values = (
            config if self.__config_connection is None else self.__config_connection
        )
        return get_str_connection(values, mode)

    def connect(self, config: dict = None):
        """
        Start oracle connection
        """
        if self.__is_connected:
            return True
        try:
            str_connection = self.__validate_config(config)
            self.__connection = cx_Oracle.connect(str_connection)
            self.__is_connected = True
            return True
        except Exception as e:
            raise e

    def close(self):
        if self.__is_connected:
            self.__connection.close()
            self.__is_connected = False

    def is_connected(self) -> bool:
        return self.__is_connected

    def get_connection(self):
        return self.__connection

    def call(
            self,
            fn: str,
            ret_type: OracleType,
            params: Optional[Dict] = None,
            custom_params: Optional[List[ZParam]] = None,
            model: Optional[models.Model] = None,
            connection=None,
            jsonfy=False,
            throw=False
    ) -> Optional[Any]:
        """
        Execute or call oracle functions - FN v0.0.1 | Core v0.0.7

        New feature for call oracle db functions
        Use this function instead function 'call'

        Parameters
        ----------
        fn : str | required
            Function name with package name: PO_LINES_PKG.FN_GET_LINE

        ret_type : OracleType | required
            The return type of oracle db function

        params : Dict | Optional
            Set parameter that the oracle funtion expects

        custom_params : Optional[List[ZParam, IntList, StrList, ClobList]] | Optional
            Custom Set parameter that the oracle funtion expects, see avaialble custom types

        model : marshmallow_objects.models.Model | Optional
            Model specification where the db data will be volcated

        connection : DB Connection | Optional
            The db connection object, if it is not passed by params, it tries to get a global instance

        jsonfy : bool | Optional
            Return data in dict format

        throw : bool | Optional
            raise exception or not
        Raises
        ------
        NotValueProvided
            Connection

        Returns
        -------
        result set : Union[List[Model],int,float,str]
            The result set of oracle db function
        """
        cursor = None
        connection_provided = True
        db_schema = None

        if connection is not None:
            cursor = connection.cursor()
        else:
            connection_provided = False
            connection = self.new_connect()
            cursor = connection.cursor()
        if connection is None:
            raise Exception("Can't get db connection")

        if self.__schemas is not None:
            db_schema = get_current_schema(self.__schemas, self.__env, self.__env)

        if custom_params is not None and len(custom_params) > 0:
            if params is None:
                params = {}
            # * Find the current env for extract the schema
            for custom in custom_params:
                params[custom.key] = make_custom_param(
                    connection,
                    param_type=custom.param_type,
                    value=custom.value,
                    schema=db_schema,
                )

        fn = (
            fn
            if db_schema is None or db_schema.replace(" ", "") == ""
            else f"{db_schema}.{fn}"
        )
        if self.__verbose:
            show_info(fn, params, ret_type, model)
        try:
            result_set = (
                cursor.callfunc(fn, ret_type.value, keywordParameters=params)
                if params is not None
                else cursor.callfunc(fn, ret_type.value)
            )
            return extract_of_result_data(result_set, ret_type, model, jsonfy)
        except Exception as e:
            process_exception(throw, e)
        finally:
            safely_exec(lambda l: l.close(), args=[cursor])
            if connection_provided is False:
                safely_exec(lambda l: l.commit(), args=[connection])
                safely_exec(lambda l: l.close(), args=[connection])

        return None

    def exec(self, fn_name: str, ret_type: OracleType, params: dict = None, list_params: List[Any] = None,
             model: Any = None,
             connection=None, jsonfy: bool = False, throw=False) -> Any:
        raise NotImplementedError("")


class ZDBOraclePool(ZDBPool):

    def __init__(self, db: ZDatabase = None, config: dict = None):
        self.db = db
        if db is None:
            self.db = ZOracle.setup(config)
        self.config = config
        self.__pool = None
        self.__max = 5
        self.__min = 1
        self.__threading = False
        self.__homogeneous = True
        self.__pool_created: bool = False
        if config is not None:
            self.setup_extras(config)

    def setup_db(self, db: ZDatabase) -> None:
        self.db = db

    def setup_extras(self, config: dict) -> None:
        self.config = config
        try:
            self.__max = config.get("max_connections", 5)
            self.__min = config.get("min_connections", 1)
            self.__threading = config.get("threading", False)
            self.__homogeneous = config.get("homogeneous", True)
        except Exception as e:
            zL.e("An error occurred when setup config", exc_info=e)

    def initialize_pool(
            self,
            max_connections: int = None,
            min_connections: int = None
    ) -> bool:
        if self.__pool_created:
            return False
        zL.i("Initializing Pool...")
        self.__pool = cx_Oracle.SessionPool(
            user=self.config.get("user"),
            password=self.config.get("password"),
            dsn=get_str_connection(self.config, mode="DSN"),
            homogeneous=self.__homogeneous,
            encoding="UTF-8",
            max=self.__max if max_connections is None else max_connections,
            min=self.__min if min_connections is None else min_connections,
            threaded=self.__threading,
        )
        self.__pool_created = True
        zL.i("Pool Started Successfully...")
        return True

    def close_pool(self, force: bool = False):
        try:
            if self.__pool_created:
                self.__pool.close(force=force)
                self.__pool_created = False
                zL.i("Pool Closed Successfully")
        except Exception as e:
            zL.e("An error occurred when try close pool", exc_info=e)

    def get_pool_connection(self):
        if self.__pool_created:
            return self.__pool.acquire()
        zL.w("DB Pool not initialized, try to initialize pool connection...")
        try:
            self.initialize_pool()
            self.__pool_created = True
            return self.__pool.acquire()
        except Exception as e:
            zL.e("Unable to initialize the connections pool...", exc_info=e)
            self.__pool_created = False

    def release_connection(self, connection) -> bool:
        try:
            if self.__pool_created:
                self.__pool.release(connection)
                return True
        except Exception as e:
            zL.e("An error occurred when try to release connection", exc_info=e)
            return False

    def get_db(self):
        return self.db



from typing import Any, Tuple

from zpy.utils.funcs import safely_exec

from zdb.commons import ZDatabase


class ZDBTransact:

    def __init__(self, db: ZDatabase, init_cursor=False):
        self.db = db
        self.session = None
        self.cursor = None
        self.init_cursor = init_cursor

    def __enter__(self):
        self.session = self.db.new_connect()
        if self.init_cursor is True:
            self.cursor = self.session.cursor()
        return self

    def run(self, statement: str, params: Tuple[Any] = None):
        if not self.init_cursor:
            raise ValueError("Current cursor is not initialized")
        self.cursor.execute(statement, params)

    def query(self, statement: str, params: Tuple[Any] = None):
        if not self.init_cursor:
            raise ValueError("Current cursor is not initialized")
        self.cursor.execute(statement, params)
        return self.cursor.fetchall()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            self.session.rollback()
        else:
            self.session.commit()
        safely_exec(lambda x: x.close(), [self.session])
        if self.init_cursor:
            safely_exec(lambda x: x.close(), [self.cursor])



from dataclasses import asdict
from typing import Optional, Any, List, Callable, TypeVar

from psycopg2 import connect, DatabaseError
from zpy.utils.funcs import safely_exec

from zdb.commons import show_info, DBTypes, get_map_type, ZDatabase, ZDBConfig, build_params, process_exception

T = TypeVar("T")


def __fn_extract_data(result_set, rows, column_names, ret_type: DBTypes, model, jsonfy=False) -> Any:
    """
        Inner Data Builder
    """
    if ret_type == DBTypes.cursor:
        if model is not None:
            return [model(**dict(zip(column_names, r))) for r in rows]
        if jsonfy is True:
            return [dict(zip(column_names, r)) for r in rows]
        return rows
    else:
        if len(rows) > 0:
            parser = get_map_type(ret_type)
            if model is not None:
                return model(**dict(zip(column_names, rows[0])))
            if jsonfy is True:
                return dict(zip(column_names, rows[0]))
            if ret_type == DBTypes.single_item:
                return rows[0]
            return parser(rows[0][0])
    return None


def extract_data(result_set, ret_type: DBTypes, cursor, model, jsonfy=False) -> Any:
    """
        Data Builder for custom stored procedures
    """
    # for result in cursor.stored_results():
    #     colum_names = result.column_names
    colum_names = [desc[0] for desc in cursor.description]
    rows = cursor.fetchall()
    if not ret_type:
        return rows
    return __fn_extract_data(result_set, rows, colum_names, ret_type, model, jsonfy)


def fn_extract_data(result_set, ret_type: DBTypes, cursor, model, jsonfy=False) -> Any:
    """
        Data Builder for custom functions
    """
    colum_names = []
    if cursor.description:
        colum_names = [desc[0] for desc in cursor.description]
    rows = cursor.fetchall()
    if not cursor.rownumber and not rows:
        return None

    if not ret_type:
        return rows
    return __fn_extract_data(result_set, rows, colum_names, ret_type, model, jsonfy)
    # print(hasattr(cursor, 'column_names'))
    # colum_names = cursor.column_names
    # rows = cursor.fetchall()
    # if not ret_type:
    #     return rows
    # return __fn_extract_data(result_set, rows, colum_names, ret_type, model, jsonfy)


class ZPostgres(ZDatabase):

    def __init__(self, dict_config: Optional[dict] = None, config: Optional[ZDBConfig] = None,
                 verbose: bool = False, auto_commit: bool = True):
        self.dict_config = dict_config
        self.config = config
        self.verbose = verbose
        self.auto_commit: bool = auto_commit

    @classmethod
    def setup(cls, config: dict, verbose: bool = False):
        return cls(config, config=None, verbose=verbose)

    @classmethod
    def setup_of(cls, config: ZDBConfig, verbose: bool = False):
        return cls(config=config, verbose=verbose)

    @classmethod
    def from_dict(cls, config: dict, verbose: bool = False, **kwargs):
        """
            dict should be contain the next keys:
            dbname=
            user=
            password=
            host=
            port=
        """
        config.update(kwargs)
        return cls(config, verbose=verbose)

    @classmethod
    def from_of(cls, user: str, password: str, host: str, db_name: str, verbose: bool = False):
        return cls({
            'user': user,
            'password': password,
            'host': host,
            'database': db_name,
            'raise_on_warnings': False
        }, verbose=verbose)

    def new_connect(self) -> Any:
        final_config = self.__validate_config()
        if final_config is None:
            raise ValueError("The value of db configuration can not be null.")
        connection = connect(**final_config)
        connection.autocommit = self.auto_commit
        return connection

    def exec(self, name: str, ret_type: Optional[DBTypes] = None, params: Optional[dict] = None,
             list_params: Optional[List[Any]] = None,
             model: Optional[Any] = None,
             connection=None, jsonfy: bool = False, throw: bool = True) -> Any:
        """Function executor

        Args:
            name (str): Stored procedure name
            ret_type (DBTypes): Type of data returned from stored procedure
            params (dict, optional): params for the procedure. Defaults to None.
            list_params (List[Any], optional): positional list params to the procedure. Defaults to None.
            model (Any, optional): model for build returned data. Defaults to None.
            connection ([type], optional): connection database. Defaults to None.
            jsonfy (bool, optional): return data in dict format if model is null. Defaults to False.
            throw (bool, optional): raise exceptions

        Returns:
            Any: processed data
        """
        return self.__call(lambda c, fn, p: c.callproc(fn, p), extract_data, name, ret_type, params, list_params, model,
                           connection, jsonfy, throw)

    def __call(self, runner: Callable, extractor: Callable, name: str, ret_type: DBTypes, params: dict = None,
               list_params: List[Any] = None,
               model: Any = None,
               connection=None, jsonfy: bool = False, throw: bool = True) -> Any:
        cn = connection
        connection_passed = True
        if connection is None:
            connection_passed = False
            cn = self.new_connect()

        cursor = None
        if cn is not None:
            cursor = cn.cursor()

        if cn is None or cursor is None:
            raise Exception("Can't get db connection")
        if self.verbose:
            show_info(name, params, ret_type, model)
        try:
            result_set = runner(cursor, name, build_params(dict_params=params, list_params=list_params))
            return extractor(result_set, ret_type, cursor, model, jsonfy)
        except DatabaseError as error:
            process_exception(throw, error)
        except Exception as e:
            process_exception(throw, e)
        finally:
            if not connection_passed:
                safely_exec(lambda c: c.close(), args=[cursor])
                if cn.closed == 0:
                    safely_exec(lambda c: c.commit(), args=[cn])
                    safely_exec(lambda c: c.close(), args=[cn])
        return None

    def call(self, fn_name: str, ret_type: DBTypes = None, params: dict = None, list_params: List[Any] = None,
             model: Any = None,
             connection=None, jsonfy: bool = False, throw: bool = True) -> Any:
        """for stored procedure

        Args:
            fn_name (str): Stored procedure name
            ret_type (DBTypes): Type of data returned from stored procedure
            params (dict, optional): params for the procedure. Defaults to None.
            list_params (List[Any], optional): positional list params to the procedure. Defaults to None.
            model (Any, optional): model for build returned data. Defaults to None.
            connection ([type], optional): connection database. Defaults to None.
            jsonfy (bool, optional): return data in dict format if model is null. Defaults to False.
            throw: (bool, optional): raise exceptions

        Returns:
            Any: processed data
        """
        return self.__call(lambda c, fn, p: c.execute(f'call {fn}', p), fn_extract_data, fn_name, ret_type, params,
                           list_params, model,
                           connection, jsonfy, throw)

    def __validate_config(self) -> dict:
        if self.config is not None:
            return asdict(self.config)
        return self.dict_config

    def init_local_client(self, path: str):
        raise NotImplementedError("Not implemented for MySQL!")

    def connect(self):
        raise NotImplementedError("Not implemented for MySQL!")

    def close(self) -> None:
        raise NotImplementedError("Not implemented for MySQL!")

    def is_connected(self) -> bool:
        raise NotImplementedError("Not implemented for MySQL!")

    def get_connection(self) -> Any:
        raise NotImplementedError("Not implemented for MySQL!")



from abc import ABC, abstractmethod
from decimal import Decimal
from enum import Enum
from typing import Union, Any, List, Dict, Callable, Optional, TypeVar
from dataclasses import dataclass

from zpy.logger import zL
from zpy.app import zapp_context as ctx

T = TypeVar('T')


def show_info(fn: str, params: Union[List[Any], Dict[Any, Any]], ret_type: Enum, v_model):
    print("\n|-------------------------------------------------|\n")
    print(f"   Function Called: {fn} ")
    print("            Params: {}".format(params))
    print("       Return Type: {}".format(ret_type.name))
    print(f"        Ref  Model: {v_model}")
    print("\n|-------------------------------------------------|\n")


@dataclass
class ZDBConfig:
    user: str
    password: str
    database: str
    host: str
    port: int = 3306
    autocommit: bool = False
    raise_on_warnings = True
    service: str = None  # Use for Oracle


parser_types = [
    None,
    lambda x: float(x),
    lambda x: str(x),
    lambda x: int(x),
    lambda x: Decimal(str(x).strip(' "')),
    lambda x: bool(x),
    lambda x: x
]


class DBTypes(Enum):
    cursor = 1
    float = 2
    string = 3
    integer = 4
    decimal = 5
    bool = 6
    single_item = 7
    out_integer = 4
    out_bool = 6


def get_map_type(r_type: DBTypes) -> Callable:
    return parser_types[r_type.value - 1]


def build_params(dict_params: dict, list_params: List[Any]) -> List[Any]:
    if dict_params is not None:
        return list(dict_params.values())
    if list_params is not None:
        return list_params
    return []


def process_exception(throw: bool, e: Exception):
    if throw is True:
        raise e
    zL.e("Failed when try to call function or stored procedure.", exc_info=e)


class ZDatabase(ABC):
    @classmethod
    @abstractmethod
    def setup(cls, config: dict, verbose: bool = False):
        """
        Setup connection arguments using dictionary
        """
        ...

    @classmethod
    @abstractmethod
    def setup_of(cls, config: ZDBConfig, verbose: bool = False):
        """
        Setup connection data
        """
        ...

    @classmethod
    @abstractmethod
    def from_of(cls, user: str, password: str, host: str, db_name: str, verbose: bool = False):
        """
        Setup connection data
        """
        ...

    def new_connect(self) -> Any:
        ...

    def __validate_config(self) -> Union[dict, str]:
        ...

    def call(self, name: str, ret_type: DBTypes, params: dict = None, list_params: List[Any] = None, model: Any = None,
             connection=None, jsonfy: bool = False, throw=False) -> Any:
        """Stored procedure caller

        Args:
            name (str): Stored procedure name
            ret_type (DBTypes): Type of data returned from stored procedure
            params (dict, optional): params for the procedure. Defaults to None.
            list_params (List[Any], optional): positional list params to the procedure. Defaults to None.
            model (Any, optional): model for build returned data. Defaults to None.
            connection ([type], optional): connection database. Defaults to None.
            jsonfy (bool, optional): return data in dict format if model is null. Defaults to False.
            throw (bool,optional)

        Returns:
            Any: processed data
        """
        ...

    def exec(self, fn_name: str, ret_type: DBTypes, params: dict = None, list_params: List[Any] = None,
             model: Any = None,
             connection=None, jsonfy: bool = False, throw=False) -> Any:
        """Function executor

        Args:
            fn_name (str): Stored procedure name
            ret_type (DBTypes): Type of data returned from stored procedure
            params (dict, optional): params for the procedure. Defaults to None.
            list_params (List[Any], optional): positional list params to the procedure. Defaults to None.
            model (Any, optional): model for build returned data. Defaults to None.
            connection ([type], optional): connection database. Defaults to None.
            jsonfy (bool, optional): return data in dcit format if model is null. Defaults to False.
            throw (bool,optional)
        Returns:
            Any: processed data
        """
        ...

    @abstractmethod
    def init_local_client(self, path: str):
        """
        Initialize local client
        """
        ...

    @abstractmethod
    def connect(self):
        ...

    @abstractmethod
    def close(self) -> None:
        ...

    @abstractmethod
    def is_connected(self) -> bool:
        ...

    @abstractmethod
    def get_connection(self) -> Any:
        ...


class ZDBPool(ABC):

    @abstractmethod
    def setup_db(self, db: ZDatabase) -> None:
        ...

    @abstractmethod
    def get_db(self):
        ...

    @abstractmethod
    def setup_extras(self, config: dict) -> None:
        ...

    @abstractmethod
    def get_pool_connection(self) -> Any:
        ...

    @abstractmethod
    def release_connection(self, connection) -> bool:
        ...

    @abstractmethod
    def close_pool(self):
        pass

    @abstractmethod
    def initialize_pool(
            self,
            max_connections: int = 5,
            min_connections: int = 1,
    ) -> bool:
        ...


@dataclass
class ZDBWrapper:
    connection: Any
    db: ZDatabase
    name: str
    verbose: bool

    def configure(self, connection: Any):
        if self.connection:
            self.release()
        self.connection = connection

    def has_connection(self) -> bool:
        return self.connection is not None

    def release(self):
        if not self.connection:
            return
        try:
            self.connection.close()
            self.connection = None
            if self.verbose:
                ctx().logger.info(f"Connection of {self.name} closed")
        except Exception as e:
            ctx().logger.err("An error occurred while releasing connections", exc_info=e)


class ZDBConnectionManager:

    def __init__(self, verbose: bool = False, auto_commit: bool = True):
        self.dbs: Dict[str, ZDBWrapper] = {}
        self.count = 0
        self.verbose = verbose
        self.auto_commit = auto_commit

    def add(self, name: str, db: ZDatabase) -> 'ZDBConnectionManager':
        """
        Add new database configuration.
        @param name: database identifier name
        @param db: database configuration instance
        @return: manager
        """
        self.dbs[name] = ZDBWrapper(None, db, name, self.verbose)
        return self

    def __configure(self, name: str):
        if name not in self.dbs:
            return
        if self.dbs.get(name).has_connection():
            ctx().logger.warn(f"Warning: Already exist a connection opened for {name} database.")
            return
        self.dbs.get(name).configure(
            self.dbs.get(name).db.new_connect()
        )
        self.count = self.count + 1
        if self.verbose:
            ctx().logger.info(f"New connection opened for {name} database.")

    def open_for(self, dbs: List[str]) -> None:
        """
        Open db connections for specified database configurations.
        @param dbs: Db names to open connections.
        @return: None
        """
        for name in dbs:
            self.__configure(name)

    def open_single(self, name: str) -> None:
        """
        Open db connections for all database configurations.
        @return: None
        """
        self.__configure(name)

    def open(self) -> None:
        """
        Open db connections for all database configurations.
        @return: None
        """
        for name in self.dbs.keys():
            self.__configure(name)

    def release_for(self, name: str) -> None:
        """
        Release specific opened connections
        @return: None
        """
        db = self.dbs.get(name, None)
        if db and db.has_connection():
            if self.auto_commit:
                self.commit_for(name)
            db.release()

    def release(self) -> None:
        """
        Release all opened connections
        @return: None
        """
        for name in self.dbs.keys():
            if self.auto_commit:
                self.commit_for(name)
            self.dbs.get(name).release()

    def count_connections(self) -> int:
        """
        Count all opened connections
        @return: total of open connections
        """
        return self.count

    def has_connections(self, name: str) -> bool:
        """
        Verify if database provided has opened connection.
        @param name: database name
        @return: true if it has connection
        """
        if name in self.dbs:
            return self.dbs.get(name).has_connection()
        return False

    def connection(self, name: str) -> [Any]:
        """
        Retrieve connection for database name provided.
        @param name: database name
        @return: connection
        """
        if name in self.dbs:
            if self.dbs.get(name).has_connection():
                return self.dbs.get(name).connection
        return None

    def commit_for(self, name: str) -> None:
        """
        Commit to specified database
        @param name: database name
        @return: connection
        """
        if name in self.dbs:
            if self.dbs.get(name).has_connection():
                self.dbs.get(name).connection.commit()

    def commit(self) -> None:
        """
        Commit all connections
        @return: connection
        """
        for db_name in self.dbs.keys():
            self.commit_for(db_name)

    def force_connection(self, name: str) -> Optional[Any]:
        """
        Force connection retrieve.
        If not existing opened connection for database provided,
        will try to open and return connection
        @param name: database name
        @return: connection
        """
        if name in self.dbs:
            if self.dbs.get(name).has_connection():
                return self.dbs.get(name).connection
            self.open_single(name)
            if self.dbs.get(name).has_connection():
                return self.dbs.get(name).connection
        return None

    def database(self, name: str, db_type: T = None) -> Union[T, Any]:
        """
        Retrieve database by name
        @param name: database name
        @param db_type: Database type
        @return:
        """
        if name in self.dbs:
            return self.dbs.get(name).db
        return None

    def __del__(self):
        self.release()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="zpy-db-core",
    version="1.2.1",
    author="Noé Cruz | linkedin.com/in/zurckz/",
    author_email="zurckz.services@gmail.com",
    description="Helper module for manage database connection",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NoeCruzMW",
    packages=setuptools.find_packages(),
    install_requires=[
        "zpy-api-core",
        "mysql-connector-python",
        "cx-Oracle",
        "psycopg2"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
